import numpy as np
import pandas as pd

import torch
import torch.nn as nn
import torch.optim as optim

from torch.utils.data import DataLoader, TensorDataset

from opacus import PrivacyEngine

from snsynth.base import Synthesizer

from ._generator import Generator
from ._discriminator import Discriminator


class DPGAN(Synthesizer):
    def __init__(
        self,
        binary=False,
        latent_dim=64,
        batch_size=64,
        epochs=1000,
        delta=None,
        epsilon=1.0,
    ):
        self.binary = binary
        self.latent_dim = latent_dim
        self.batch_size = batch_size
        self.epochs = epochs
        self.delta = delta
        self.epsilon = epsilon

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        self.pd_cols = None
        self.pd_index = None

    def train(
        self,
        data,
        categorical_columns=None,
        ordinal_columns=None,
        update_epsilon=None,
        transformer=None,
        continuous_columns=None,
        preprocessor_eps=0.0,
        nullable=False,
    ):
        if update_epsilon:
            self.epsilon = update_epsilon

        train_data = self._get_train_data(
            data,
            style='gan',
            transformer=transformer,
            categorical_columns=categorical_columns, 
            ordinal_columns=ordinal_columns, 
            continuous_columns=continuous_columns, 
            nullable=nullable,
            preprocessor_eps=preprocessor_eps
        )

        data = np.array(train_data)

        if isinstance(data, pd.DataFrame):
            for col in data.columns:
                data[col] = pd.to_numeric(data[col], errors="ignore")
            self.pd_cols = data.columns
            self.pd_index = data.index
            data = data.to_numpy()
        elif isinstance(data, list):
            data = np.array(data)
        elif not isinstance(data, np.ndarray):
            raise ValueError("Data must be a numpy array or pandas dataframe")

        dataset = TensorDataset(
            torch.from_numpy(data.astype("float32")).to(self.device)
        )
        dataloader = DataLoader(
            dataset, batch_size=self.batch_size, shuffle=True, drop_last=True
        )

        self.generator = Generator(
            self.latent_dim, data.shape[1], binary=self.binary
        ).to(self.device)
        discriminator = Discriminator(data.shape[1]).to(self.device)
        optimizer_d = optim.Adam(discriminator.parameters(), lr=4e-4)

        privacy_engine = PrivacyEngine(
            discriminator,
            batch_size=self.batch_size,
            sample_size=len(data),
            alphas=[1 + x / 10.0 for x in range(1, 100)] + list(range(12, 64)),
            noise_multiplier=3.5,
            max_grad_norm=1.0,
            clip_per_layer=True,
        )

        privacy_engine.attach(optimizer_d)
        optimizer_g = optim.Adam(self.generator.parameters(), lr=1e-4)

        criterion = nn.BCELoss()

        if self.delta is None:
            self.delta = 1 / (data.shape[0] * np.sqrt(data.shape[0]))

        for epoch in range(self.epochs):
            eps, best_alpha = optimizer_d.privacy_engine.get_privacy_spent(self.delta)

            if self.epsilon < eps:
                if epoch == 0:
                    raise ValueError(
                        "Inputted epsilon and sigma parameters are too small to"
                        + " create a private dataset. Try increasing either parameter and rerunning."
                    )
                break

            for i, data in enumerate(dataloader):
                discriminator.zero_grad()

                real_data = data[0].to(self.device)

                # train with fake data
                noise = torch.randn(
                    self.batch_size, self.latent_dim, 1, 1, device=self.device
                )
                noise = noise.view(-1, self.latent_dim)
                fake_data = self.generator(noise)
                label_fake = torch.full(
                    (self.batch_size,), 0, dtype=torch.float, device=self.device
                )
                output = discriminator(fake_data.detach())
                loss_d_fake = criterion(output.squeeze(), label_fake)
                loss_d_fake.backward()
                optimizer_d.step()

                # train with real data
                label_true = torch.full(
                    (self.batch_size,), 1, dtype=torch.float, device=self.device
                )
                output = discriminator(real_data.float())
                loss_d_real = criterion(output.squeeze(), label_true)
                loss_d_real.backward()
                optimizer_d.step()

                max_grad_norm = []
                for p in discriminator.parameters():
                    param_norm = p.grad.data.norm(2).item()
                    max_grad_norm.append(param_norm)

                privacy_engine.max_grad_norm = max_grad_norm

                # train generator
                self.generator.zero_grad()
                label_g = torch.full(
                    (self.batch_size,), 1, dtype=torch.float, device=self.device
                )
                output_g = discriminator(fake_data)
                loss_g = criterion(output_g.squeeze(), label_g)
                loss_g.backward()
                optimizer_g.step()

                # manually clear gradients
                for p in discriminator.parameters():
                    if hasattr(p, "grad_sample"):
                        del p.grad_sample
                # autograd_grad_sample.clear_backprops(discriminator)

                if self.delta is None:
                    self.delta = 1 / data.shape[0]

    def generate(self, n):
        steps = n // self.batch_size + 1
        data = []
        for i in range(steps):
            noise = torch.randn(
                self.batch_size, self.latent_dim, 1, 1, device=self.device
            )
            noise = noise.view(-1, self.latent_dim)

            fake_data = self.generator(noise)
            data.append(fake_data.detach().cpu().numpy())

        data = np.concatenate(data, axis=0)
        data = data[:n]

        return self._transformer.inverse_transform(data)

    def fit(self, data, *ignore, transformer=None, categorical_columns=[], ordinal_columns=[], continuous_columns=[], preprocessor_eps=0.0, nullable=False):
        self.train(data, transformer=transformer, categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, preprocessor_eps=preprocessor_eps, nullable=nullable)

    def sample(self, n_samples):
        return self.generate(n_samples)



from functools import wraps

import numpy as np
import pandas as pd
import warnings

from snsynth.base import SDGYMBaseSynthesizer


class PytorchDPSynthesizer(SDGYMBaseSynthesizer):
    def __init__(self, epsilon, gan, preprocessor=None):
        """
        Wrapper class to unify pytorch GAN architectures with the SDGYM API.

        :param epsilon: Total epsilon used for the DP Synthesizer
        :type epsilon: float
        :param gan: A pytorch defined GAN
        :type gan: torch.nn.Module
        :param preprocessor: A preprocessor to .transform the input data and
            .inverse_transform the output of the GAN., defaults to None
        :type preprocessor: GeneralTransformer, optional
        """
        self.epsilon = epsilon
        self.gan = gan
        self.preprocessor = preprocessor

        self._data_columns = None

    def _get_training_data(self, data, categorical_columns, ordinal_columns):
        if not self.preprocessor:
            return data
        else:
            self.preprocessor.fit(data, categorical_columns, ordinal_columns)
            return self.preprocessor.transform(data)

    @wraps(SDGYMBaseSynthesizer.fit)
    def fit(
        self,
        data,
        categorical_columns=tuple(),
        ordinal_columns=tuple(),
        transformer=None,
        continuous_columns=None,
        preprocessor_eps=0.0,
        nullable=False,
    ):
        def column_names(n_items, prefix="col"):
            names = []
            for i in range(n_items):
                names.append(prefix + "_" + str(i))
            return names

        if isinstance(data, pd.DataFrame):
            self._data_columns = data.columns
        elif isinstance(data, np.ndarray):
            placeholder_columns = column_names(data.shape[1])
            data = pd.DataFrame(data, columns=placeholder_columns).infer_objects()
            self._data_columns = placeholder_columns
            warnings.warn(
                "Data is numpy array, converting to pandas dataframe with default "
                + "column names. Inferring data types. Note: for best performance, "
                + "pandas dataframe should be constructed by user and "
                + "data_types should be specified beforehand. Dtypes: "
                + str(data.dtypes),
                Warning,
            )

        self.dtypes = data.dtypes

        training_data = self._get_training_data(
            data, categorical_columns, ordinal_columns
        )

        self.gan.train(
            training_data,
            categorical_columns=categorical_columns,
            ordinal_columns=ordinal_columns,
            update_epsilon=self.epsilon,
            transformer=transformer,
            continuous_columns=continuous_columns,
            preprocessor_eps=preprocessor_eps,
            nullable=nullable,
        )

    @wraps(SDGYMBaseSynthesizer.sample)
    def sample(self, n):
        synth_data = self.gan.generate(n)

        if self.preprocessor is not None:
            synth_data = self.preprocessor.inverse_transform(synth_data)

        if isinstance(synth_data, np.ndarray):
            synth_data = pd.DataFrame(synth_data, columns=self._data_columns)
        elif isinstance(synth_data, pd.DataFrame):
            # TODO: Add validity check
            synth_data.columns = self._data_columns
        else:
            raise ValueError("Generated data is neither numpy array nor dataframe!")

        return synth_data



from .table import TableTransformer
from .onehot import OneHotEncoder
from .label import LabelTransformer
from .minmax import MinMaxTransformer
from .bin import BinTransformer
from .chain import ChainTransformer
from .log import LogTransformer
from .standard import StandardScaler
from .clamp import ClampTransformer
from .log import LogTransformer
from .table import NoTransformer
from .anonymization import AnonymizationTransformer
from .drop import DropTransformer
__all__ = [
    "TableTransformer", 
    "OneHotEncoder", 
    "LabelTransformer", 
    "MinMaxTransformer", 
    "BinTransformer", 
    "ChainTransformer", 
    "LogTransformer",
    "StandardScaler",
    "LogTransformer",
    "ClampTransformer",
    "NoTransformer",
    "AnonymizationTransformer",
    "DropTransformer"
    ]


import pandas as pd
from sklearn.base import RegressorMixin, MultiOutputMixin
from sklearn.linear_model.base import LinearModel
from .dp_covariance import DPcovariance


class DPLinearRegression(MultiOutputMixin, RegressorMixin, LinearModel):
    """
    Differentially Private Linear Regression using the Covariance Method.

    Fits a linear model with coefficients w = (w1, ..., wp)
    to minimize the residual sum of squares between the observed targets in
    the dataset, and the targets predicted by the linear approximation.

    Parameters
    ----------
    fit_intercept : bool, optional, default True
        Whether to calculate the intercept for this model. If set
        to False, no intercept will be used in calculations
        (i.e. data is expected to be centered).

    Attributes
    ----------
    coef_ : array of shape (n_features, ) or (n_targets, n_features)
        Estimated coefficients for the linear regression problem.
        If multiple targets are passed during the fit (y 2D), this
        is a 2D array of shape (n_targets, n_features), while if only
        one target is passed, this is a 1D array of length n_features.

    intercept_ : float or array of shape of (n_targets,)
        Independent term in the linear model. Set to 0.0 if
        `fit_intercept = False`.



    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.linear_model import LinearRegression
    >>> X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
    >>> # y = 1 * x_0 + 2 * x_1 + 3
    >>> y = np.dot(X, np.array([1, 2])) + 3
    >>> reg = LinearRegression().fit(X, y)
    >>> reg.score(X, y)
    1.0
    >>> reg.coef_
    array([1., 2.])
    >>> reg.intercept_
    3.0000...
    >>> reg.predict(np.array([[3, 5]]))
    array([16.])
    """

    def __init__(self, fit_intercept=False):
        self.fit_intercept = fit_intercept

    def _set_coef_and_intercept(self, df):
        df_copy = df
        if self.fit_intercept:
            self.intercept_ = df_copy["intercept", "Estimate"]
            df_copy = df_copy.drop(["intercept"])
        self.coef_ = df_copy["Estimate"].values

    def fit(self, X, y, bounds, budget):
        """
        Fit linear model.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training data

        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            Target values. Will be cast to X's dtype if necessary


        Returns
        -------
        self : returns an instance of self.
        """
        data = pd.concat((X, y), axis=1)
        n = data.shape[0]

        cols = list(X.columns.values) + list(y.columns.values)

        results = DPcovariance(n, cols, bounds, budget).get_linear_regression(
            data, X.columns.values, y.columns.values, self.fit_intercept
        )
        self._set_coef_and_intercept(results)
        return self



from .base import ColumnTransformer
from .definitions import ColumnType


class DropTransformer(ColumnTransformer):
    """
    Transformer that ignores a column completely. All values will be dropped during transformation.
    Inverse transformation is a no-op.
    """
    def __init__(self):
        super().__init__()

    @property
    def output_type(self):
        return ColumnType.UNBOUNDED

    @property
    def cardinality(self):
        return [None]

    def _fit(self, _):
        pass

    def _clear_fit(self):
        self._fit_complete = True
        self.output_width = 0

    def _transform(self, _):
        pass

    def _inverse_transform(self, _):
        pass

    def transform(self, data, idx=None):
        if idx is None:
            return [None for _ in data]
        else:
            return [row[:idx] + row[idx + 1 :] for row in data]

    def inverse_transform(self, data, idx=None):
        return data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['snsynth',
 'snsynth.aggregate_seeded',
 'snsynth.models',
 'snsynth.mst',
 'snsynth.pytorch',
 'snsynth.pytorch.nn',
 'snsynth.pytorch.nn.ctgan',
 'snsynth.transform']

package_data = \
{'': ['*']}

install_requires = \
['Faker>=15.0.0,<16.0.0',
 'opacus>=0.14.0,<0.15.0',
 'opendp>=0.6.0,<0.7.0',
 'pac-synth>=0.0.6,<0.0.7',
 'smartnoise-sql>=0.2.6,<0.3.0']

setup_kwargs = {
    'name': 'smartnoise-synth',
    'version': '0.3.5',
    'description': 'Differentially Private Synthetic Data',
    'long_description': '[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT) [![Python](https://img.shields.io/badge/python-3.7%20%7C%203.8-blue)](https://www.python.org/)\n\n<a href="https://smartnoise.org"><img src="https://github.com/opendp/smartnoise-sdk/raw/main/images/SmartNoise/SVG/Logo%20Mark_grey.svg" align="left" height="65" vspace="8" hspace="18"></a>\n\n# SmartNoise Synthesizers\n\nDifferentially private synthesizers for tabular data.  Package includes:\n* MWEM\n* MST\n* QUAIL\n* DP-CTGAN\n* PATE-CTGAN\n* PATE-GAN\n\n## Installation\n\n```\npip install smartnoise-synth\n```\n\n## Using\n\nPlease see the [SmartNoise synthesizers documentation](https://docs.smartnoise.org/synth/index.html) for usage examples.\n\n## Note on Inputs\n\nMWEM and MST require columns to be categorical. If you have columns with continuous values, you should discretize them before fitting.  Take care to discretize in a way that does not reveal information about the distribution of the data.\n\n## Communication\n\n- You are encouraged to join us on [GitHub Discussions](https://github.com/opendp/opendp/discussions/categories/smartnoise)\n- Please use [GitHub Issues](https://github.com/opendp/smartnoise-sdk/issues) for bug reports and feature requests.\n- For other requests, including security issues, please contact us at [smartnoise@opendp.org](mailto:smartnoise@opendp.org).\n\n## Releases and Contributing\n\nPlease let us know if you encounter a bug by [creating an issue](https://github.com/opendp/smartnoise-sdk/issues).\n\nWe appreciate all contributions. Please review the [contributors guide](../contributing.rst). We welcome pull requests with bug-fixes without prior discussion.\n\nIf you plan to contribute new features, utility functions or extensions to this system, please first open an issue and discuss the feature with us.\n',
    'author': 'SmartNoise Team',
    'author_email': 'smartnoise@opendp.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://smartnoise.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



from collections import namedtuple
import numpy as np
import torch
from torch import optim
from torch import nn
import torch.utils.data
from torch.nn import (
    BatchNorm1d,
    Dropout,
    LeakyReLU,
    Linear,
    Module,
    ReLU,
    Sequential,
    Sigmoid,
)
import warnings

import opacus
from snsynth.base import Synthesizer

from snsynth.transform.table import TableTransformer
from .ctgan.data_sampler import DataSampler
from .ctgan.ctgan import CTGANSynthesizer


class Discriminator(Module):
    def __init__(self, input_dim, discriminator_dim, loss, pac=1):
        super(Discriminator, self).__init__()
        torch.cuda.manual_seed(0)
        torch.manual_seed(0)

        dim = input_dim * pac
        #  print ('now dim is {}'.format(dim))
        self.pac = pac
        self.pacdim = dim

        seq = []
        for item in list(discriminator_dim):
            seq += [Linear(dim, item), LeakyReLU(0.2), Dropout(0.5)]
            dim = item

        seq += [Linear(dim, 1)]
        if loss == "cross_entropy":
            seq += [Sigmoid()]
        self.seq = Sequential(*seq)

    def calc_gradient_penalty(
        self, real_data, fake_data, device="cpu", pac=1, lambda_=10
    ):
        alpha = torch.rand(real_data.size(0) // pac, 1, 1, device=device)
        alpha = alpha.repeat(1, pac, real_data.size(1))
        alpha = alpha.view(-1, real_data.size(1))

        interpolates = alpha * real_data + ((1 - alpha) * fake_data)

        disc_interpolates = self(interpolates)

        gradients = torch.autograd.grad(
            outputs=disc_interpolates,
            inputs=interpolates,
            grad_outputs=torch.ones(disc_interpolates.size(), device=device),
            create_graph=True,
            retain_graph=True,
            only_inputs=True,
        )[0]

        gradient_penalty = (
            (gradients.view(-1, pac * real_data.size(1)).norm(2, dim=1) - 1) ** 2
        ).mean() * lambda_

        return gradient_penalty

    def forward(self, input):
        assert input.size()[0] % self.pac == 0
        return self.seq(input.view(-1, self.pacdim))


class Residual(Module):
    def __init__(self, i, o):
        super(Residual, self).__init__()
        self.fc = Linear(i, o)
        self.bn = BatchNorm1d(o)
        self.relu = ReLU()

    def forward(self, input):
        out = self.fc(input)
        out = self.bn(out)
        out = self.relu(out)
        return torch.cat([out, input], dim=1)


class Generator(Module):
    def __init__(self, embedding_dim, generator_dim, data_dim):
        super(Generator, self).__init__()
        dim = embedding_dim
        seq = []
        for item in list(generator_dim):
            seq += [Residual(dim, item)]
            dim += item
        seq.append(Linear(dim, data_dim))
        self.seq = Sequential(*seq)

    def forward(self, input):
        data = self.seq(input)
        return data


# custom for calcuate grad_sample for multiple loss.backward()
def _custom_create_or_extend_grad_sample(
    param: torch.Tensor, grad_sample: torch.Tensor, batch_dim: int
) -> None:
    """
    Create a 'grad_sample' attribute in the given parameter, or accumulate it
    if the 'grad_sample' attribute already exists.
    This custom code will not work when using optimizer.virtual_step()
    """

    if hasattr(param, "grad_sample"):
        param.grad_sample = param.grad_sample + grad_sample
        # param.grad_sample = torch.cat((param.grad_sample, grad_sample), batch_dim)
    else:
        param.grad_sample = grad_sample


class DPCTGAN(CTGANSynthesizer, Synthesizer):
    """DPCTGAN Synthesizer.

    GAN-based synthesizer that uses conditional masks to learn tabular data.

    :param epsilon: Privacy budget for the model.
    :param sigma: The noise scale for the gradients.  Noise scale and batch size influence
        how fast the privacy budget is consumed, which in turn influences the
        convergence rate and the quality of the synthetic data.
    :param batch_size: The batch size for training the model.
    :param epochs: The number of epochs to train the model.
    :param embedding_dim: The dimensionality of the embedding layer.
    :param generator_dim: The dimensionality of the generator layer.
    :param discriminator_dim: The dimensionality of the discriminator layer.
    :param generator_lr: The learning rate for the generator.
    :param discriminator_lr: The learning rate for the discriminator.
    :param verbose: Whether to print the training progress.
    :param diabled_dp: Allows training without differential privacy, to diagnose
        whether any model issues are caused by privacy or are simply the
        result of GAN instability or other issues with hyperparameters.

    """
    def __init__(
        self,
        embedding_dim=128,
        generator_dim=(256, 256),
        discriminator_dim=(256, 256),
        generator_lr=2e-4,
        generator_decay=1e-6,
        discriminator_lr=2e-4,
        discriminator_decay=1e-6,
        batch_size=500,
        discriminator_steps=1,
        verbose=True,
        epochs=300,
        pac=1,
        cuda=True,
        disabled_dp=False,
        delta=None,
        sigma=5,
        max_per_sample_grad_norm=1.0,
        epsilon=1,
        loss="cross_entropy"
    ):

        assert batch_size % 2 == 0

        self._embedding_dim = embedding_dim
        self._generator_dim = generator_dim
        self._discriminator_dim = discriminator_dim

        self._generator_lr = generator_lr
        self._generator_decay = generator_decay
        self._discriminator_lr = discriminator_lr
        self._discriminator_decay = discriminator_decay

        self._batch_size = batch_size
        self._discriminator_steps = discriminator_steps
        self._verbose = verbose
        self._epochs = epochs
        self.pac = pac

        # opacus parameters
        self.sigma = sigma
        self.disabled_dp = disabled_dp
        self.delta = delta
        self.max_per_sample_grad_norm = max_per_sample_grad_norm
        self.epsilon = epsilon
        self.epsilon_list = []
        self.alpha_list = []
        self.loss_d_list = []
        self.loss_g_list = []
        self.verbose = verbose
        self.loss = loss

        if not cuda or not torch.cuda.is_available():
            device = "cpu"
        elif isinstance(cuda, str):
            device = cuda
        else:
            device = "cuda"

        self._device = torch.device(device)

        self._transformer = None
        self._data_sampler = None
        self._generator = None

        if self.loss != "cross_entropy":
            # Monkeypatches the _create_or_extend_grad_sample function when calling opacus
            opacus.grad_sample.utils.create_or_extend_grad_sample = (
                _custom_create_or_extend_grad_sample
            )

    def train(
        self,
        data,
        transformer=None,
        categorical_columns=[],
        ordinal_columns=[],
        continuous_columns=[],
        update_epsilon=None, 
        preprocessor_eps=0.0,
        nullable=False,
    ):
        if update_epsilon:
            self.epsilon = update_epsilon

        train_data = self._get_train_data(
            data,
            style='gan',
            transformer=transformer,
            categorical_columns=categorical_columns, 
            ordinal_columns=ordinal_columns, 
            continuous_columns=continuous_columns, 
            nullable=nullable,
            preprocessor_eps=preprocessor_eps
        )

        train_data = np.array([
            [float(x) if x is not None else 0.0 for x in row] for row in train_data
        ])

        self._data_sampler = DataSampler(
            train_data,
            self._transformer.transformers
        )

        data_dim = self._transformer.output_width

        self._generator = Generator(
            self._embedding_dim + self._data_sampler.dim_cond_vec(),
            self._generator_dim,
            data_dim,
        ).to(self._device)

        discriminator = Discriminator(
            data_dim + self._data_sampler.dim_cond_vec(),
            self._discriminator_dim,
            self.loss,
            self.pac,
        ).to(self._device)

        optimizerG = optim.Adam(
            self._generator.parameters(),
            lr=self._generator_lr,
            betas=(0.5, 0.9),
            weight_decay=self._generator_decay,
        )
        optimizerD = optim.Adam(
            discriminator.parameters(),
            lr=self._discriminator_lr,
            betas=(0.5, 0.9),
            weight_decay=self._discriminator_decay,
        )

        privacy_engine = opacus.PrivacyEngine(
            discriminator,
            batch_size=self._batch_size,
            sample_size=train_data.shape[0],
            alphas=[1 + x / 10.0 for x in range(1, 100)] + list(range(12, 64)),
            noise_multiplier=self.sigma,
            max_grad_norm=self.max_per_sample_grad_norm,
            clip_per_layer=True,
        )

        if not self.disabled_dp:
            privacy_engine.attach(optimizerD)

        one = torch.tensor(1, dtype=torch.float).to(self._device)
        mone = one * -1

        real_label = 1
        fake_label = 0
        criterion = nn.BCELoss()

        assert self._batch_size % 2 == 0
        mean = torch.zeros(self._batch_size, self._embedding_dim, device=self._device)
        std = mean + 1

        steps_per_epoch = max(len(train_data) // self._batch_size, 1)
        for i in range(self._epochs):
            if not self.disabled_dp:
                # if self.loss == 'cross_entropy':
                #    autograd_grad_sample.clear_backprops(discriminator)
                # else:
                for p in discriminator.parameters():
                    if hasattr(p, "grad_sample"):
                        del p.grad_sample

                if self.delta is None:
                    self.delta = 1 / (
                        train_data.shape[0] * np.sqrt(train_data.shape[0])
                    )

                epsilon, best_alpha = optimizerD.privacy_engine.get_privacy_spent(
                    self.delta
                )

                self.epsilon_list.append(epsilon)
                self.alpha_list.append(best_alpha)
                if self.epsilon < epsilon:
                    if self._epochs == 1:
                        raise ValueError(
                            "Inputted epsilon and sigma parameters are too small to"
                            + " create a private dataset. Try increasing either parameter "
                            + "and rerunning."
                        )
                    else:
                        break

            for id_ in range(steps_per_epoch):
                fakez = torch.normal(mean=mean, std=std)

                condvec = self._data_sampler.sample_condvec(self._batch_size)
                if condvec is None:
                    c1, m1, col, opt = None, None, None, None
                    real = self._data_sampler.sample_data(self._batch_size, col, opt)
                else:
                    c1, m1, col, opt = condvec
                    c1 = torch.from_numpy(c1).to(self._device)
                    m1 = torch.from_numpy(m1).to(self._device)
                    fakez = torch.cat([fakez, c1], dim=1)

                    perm = np.arange(self._batch_size)
                    np.random.shuffle(perm)
                    real = self._data_sampler.sample_data(
                        self._batch_size, col[perm], opt[perm]
                    )
                    c2 = c1[perm]

                fake = self._generator(fakez)
                fakeact = self._apply_activate(fake)

                real = torch.from_numpy(real.astype("float32")).to(self._device)

                if c1 is not None:
                    fake_cat = torch.cat([fakeact, c1], dim=1)
                    real_cat = torch.cat([real, c2], dim=1)
                else:
                    real_cat = real
                    fake_cat = fakeact

                optimizerD.zero_grad()

                if self.loss == "cross_entropy":
                    y_fake = discriminator(fake_cat)

                    #   print ('y_fake is {}'.format(y_fake))
                    label_fake = torch.full(
                        (int(self._batch_size / self.pac),),
                        fake_label,
                        dtype=torch.float,
                        device=self._device,
                    )

                    #    print ('label_fake is {}'.format(label_fake))

                    error_d_fake = criterion(y_fake.squeeze(), label_fake)
                    error_d_fake.backward()
                    optimizerD.step()

                    # train with real
                    label_true = torch.full(
                        (int(self._batch_size / self.pac),),
                        real_label,
                        dtype=torch.float,
                        device=self._device,
                    )
                    y_real = discriminator(real_cat)
                    error_d_real = criterion(y_real.squeeze(), label_true)
                    error_d_real.backward()
                    optimizerD.step()

                    loss_d = error_d_real + error_d_fake

                else:

                    y_fake = discriminator(fake_cat)
                    mean_fake = torch.mean(y_fake)
                    mean_fake.backward(one)

                    y_real = discriminator(real_cat)
                    mean_real = torch.mean(y_real)
                    mean_real.backward(mone)

                    optimizerD.step()

                    loss_d = -(mean_real - mean_fake)

                max_grad_norm = []
                for p in discriminator.parameters():
                    param_norm = p.grad.data.norm(2).item()
                    max_grad_norm.append(param_norm)
                # pen = calc_gradient_penalty(discriminator, real_cat, fake_cat, self.device)

                # pen.backward(retain_graph=True)
                # loss_d.backward()
                # optimizer_d.step()

                fakez = torch.normal(mean=mean, std=std)
                condvec = self._data_sampler.sample_condvec(self._batch_size)

                if condvec is None:
                    c1, m1, col, opt = None, None, None, None
                else:
                    c1, m1, col, opt = condvec
                    c1 = torch.from_numpy(c1).to(self._device)
                    m1 = torch.from_numpy(m1).to(self._device)
                    fakez = torch.cat([fakez, c1], dim=1)

                fake = self._generator(fakez)
                fakeact = self._apply_activate(fake)

                if c1 is not None:
                    y_fake = discriminator(torch.cat([fakeact, c1], dim=1))
                else:
                    y_fake = discriminator(fakeact)

                # if condvec is None:
                cross_entropy = 0
                # else:
                #    cross_entropy = self._cond_loss(fake, c1, m1)

                if self.loss == "cross_entropy":
                    label_g = torch.full(
                        (int(self._batch_size / self.pac),),
                        real_label,
                        dtype=torch.float,
                        device=self._device,
                    )
                    # label_g = torch.full(int(self.batch_size/self.pack,),1,device=self.device)
                    loss_g = criterion(y_fake.squeeze(), label_g)
                    loss_g = loss_g + cross_entropy
                else:
                    loss_g = -torch.mean(y_fake) + cross_entropy

                optimizerG.zero_grad()
                loss_g.backward()
                optimizerG.step()

            self.loss_d_list.append(loss_d)
            self.loss_g_list.append(loss_g)
            if self.verbose:
                print(
                    "Epoch %d, Loss G: %.4f, Loss D: %.4f"
                    % (i + 1, loss_g.detach().cpu(), loss_d.detach().cpu()),
                    flush=True,
                )
                print("epsilon is {e}, alpha is {a}".format(e=epsilon, a=best_alpha))

        return self.loss_d_list, self.loss_g_list, self.epsilon_list, self.alpha_list

    def generate(self, n, condition_column=None, condition_value=None):
        """
        TODO: Add condition_column support from CTGAN
        """
        self._generator.eval()

        # output_info = self._transformer.output_info
        steps = n // self._batch_size + 1
        data = []
        for i in range(steps):
            mean = torch.zeros(self._batch_size, self._embedding_dim)
            std = mean + 1
            fakez = torch.normal(mean=mean, std=std).to(self._device)

            condvec = self._data_sampler.sample_condvec(self._batch_size)

            if condvec is None:
                pass
            else:
                c1, m1, col, opt = condvec
                c1 = torch.from_numpy(c1).to(self._device)
                fakez = torch.cat([fakez, c1], dim=1)

            fake = self._generator(fakez)
            fakeact = self._apply_activate(fake)
            data.append(fakeact.detach().cpu().numpy())

        data = np.concatenate(data, axis=0)
        data = data[:n]

        return self._transformer.inverse_transform(data)

    def fit(self, data, *ignore, transformer=None, categorical_columns=[], ordinal_columns=[], continuous_columns=[], preprocessor_eps=0.0, nullable=False):
        self.train(data, transformer=transformer, categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, preprocessor_eps=preprocessor_eps, nullable=nullable)

    def sample(self, n_samples):
        return self.generate(n_samples)



import math
import random
from typing import List
import warnings
from itertools import combinations, product

import numpy as np
import pandas as pd

from snsynth.base import Synthesizer

class Query:
    def __init__(self, query):
        self.query = query

    def evaluate(self, hist):
        e = hist.T[tuple(self.query)]
        if isinstance(e, np.ndarray):
            return np.sum(e)
        else:
            return e

    def error(self, hist, synth_hist):
        return np.abs(self.evaluate(hist) - self.evaluate(synth_hist))

    def mask(self, hist):
        data = np.zeros_like(hist.copy())
        view = data.copy()
        view.T[tuple(self.query)] = 1.0
        return view

    @property
    def queries(self):
        return [self]

    @classmethod
    def make_arbitrary(cls, dims):
        inds = []
        for _, s in np.ndenumerate(dims):
            # Random linear sample, within dimensions
            a = np.random.randint(s)
            b = np.random.randint(s)
            l_b = min(a, b)
            u_b = max(a, b) + 1
            pre = []
            pre.append(l_b)
            pre.append(u_b)
            inds.append(pre)
        # Compose slice
        sl = []
        for ind in inds:
            sl.append(np.s_[ind[0]: ind[1]])
        return Query(sl)

    @classmethod
    def make_marginals(cls, dims_mask):
        # Makes all marginal slices matching the dimensions
        # Some should be set to None to marginalize.  For example,
        # if dims = (2,None,7), then the middle dimension is marginalized,
        # and slices for all of 2x7 are made.
        dims_mask = list(reversed(dims_mask))
        mask = [1 if v is not None else 0 for v in dims_mask]
        mask = np.cumsum(mask) - 1
        mask = [x if y is not None else None for x, y in zip(mask, dims_mask)]

        dims = [d for d in dims_mask if d is not None]
        ranges = [range(d) for d in dims]
        prod = product(*ranges)  # cartesian product

        return [
            Query([
                np.s_[:] if v is None else np.s_[p[v]] for v in mask
            ]) for p in prod
        ]


class Cuboid:
    # A cuboid is a collection of marginal queries that are mutually disjoint
    def __init__(self, queries, dims_mask):
        self.queries = queries
        dims_mask = dims_mask

    @property
    def n_cells(self):
        return len(self.queries)

    @property
    def n_cols(self):
        return np.sum([1 if c is not None else 0 for c in self.dims_mask])

    def error(self, hist, synth_hist):
        err = np.sum([q.error(hist, synth_hist) for q in self.queries])
        err = err / self.n_cells
        return err if err > 0.0 else 0.0

    @classmethod
    def make_cuboid(cls, dimensions, columns):
        assert max(columns) < len(dimensions)
        dims_mask = [d if c in columns else None for c, d in enumerate(dimensions)]
        queries = Query.make_marginals(dims_mask)
        return Cuboid(queries, dims_mask)

    @classmethod
    def make_n_way(cls, dimensions, n):
        n_dims = len(dimensions)
        if n_dims < n:
            return []
        indices = list(np.arange(n_dims))
        combos = list(combinations(indices, n))
        return [cls.make_cuboid(dimensions, c) for c in combos]


class Histogram:
    def __init__(self, data, dimensions, bins, split):
        self.data = data
        self.dimensions = dimensions
        self.bins = bins
        self.split = split
        self.queries = []
        assert (len(self.dimensions) == len(self.bins))
        assert (len(self.dimensions) == len(self.split))
        assert (all([a == b for a, b in zip(self.data.shape, self.dimensions)]))

    @property
    def dimensionality(self):
        return np.prod(self.dimensions)

    @property
    def n_cols(self):
        return len(self.dimensions)

    @property
    def n_cuboids(self):
        return 2 ** self.n_cols - 1

    @property
    def n_queries(self):
        return len(self.queries)

    @property
    def n_slices(self):
        return np.sum([len(q.queries) for q in self.queries])

    def add_arbitrary_queries(self, n_queries):
        for _ in range(n_queries):
            self.queries.append(Query.make_arbitrary(self.dimensions))

    def add_marginal_queries(self, max_cols=2):
        dims = self.dimensions
        for n in range(1, max_cols + 1):
            cuboids = Cuboid.make_n_way(dims, n)
            for c in cuboids:
                self.queries.extend(c.queries)

    def add_cuboid_queries(self, max_cols):
        dims = self.dimensions
        max_cols = max_cols if max_cols <= len(dims) else len(dims)
        for n in range(1, max_cols + 1):
            cuboids = Cuboid.make_n_way(dims, n)
            self.queries.extend(cuboids)

    @classmethod
    def histogramdd_indexes(cls, x: np.ndarray, category_lengths: List[int]) -> np.ndarray:
        # https://github.com/opendp/prelim/blob/main/python/stat_histogram.py#L9-L31
        """Compute counts of each combination of categories in d dimensions.
        Discrete version of np.histogramdd.
        :param x: data of shape [n, len(`category_lengths`)] of non-negative category indexes
        :param category_lengths: the number of unique categories per column
        """

        assert x.shape[1] == len(category_lengths)
        assert x.ndim == 2
        if not len(category_lengths):
            return np.array(x.shape[0])

        # consider each row as a multidimensional index into an ndarray
        # determine what those indexes would be should the ndarray be flattened
        # the flat indices uniquely identify each cell
        flat_indices = np.ravel_multi_index(x.T, category_lengths)

        # count the number of instances of each index
        hist = np.bincount(flat_indices, minlength=np.prod(category_lengths))

        # map counts back to d-dimensional output
        return hist.reshape(category_lengths)


class MWEMSynthesizer(Synthesizer):
    """
        N-Dimensional numpy implementation of MWEM.
        (http://users.cms.caltech.edu/~katrina/papers/mwem-nips.pdf)

    From the paper:
    "[MWEM is] a broadly applicable, simple, and easy-to-implement
    algorithm, capable of substantially improving the performance of
    linear queries on many realistic datasets...
    (circa 2012)...MWEM matches the best known and nearly
    optimal theoretical accuracy guarantees for differentially private
    data analysis with linear queries."

    Linear queries used for sampling in this implementation are
    random contiguous slices of the n-dimensional numpy array.

    :param epsilon: Privacy budget.
    :type epsilon: float, optional
    :param q_count: Number of random queries in the pool to generate.
        Must be more than # of iterations
    :type q_count: int, optional
    :param iterations: Number of iterations of MWEM to run.  MWEM will
        guess a reasonable number of iterations if this is not specified.
    :type iterations: int, optional
    :param splits: Allows you to specify feature dependence when creating
        internal histograms.
        Columns that are known to be dependent can be kept together.
        Example: splits=[[0,1],[2,3]] where
        columns 0 and 1 are dependent, columns 2 and 3 are dependent,
        and between groupings there is independence, defaults to []
    :type splits: list, optional
    :param split_factor: If splits not specified, can instead subdivide
        pseudo-randomly. For example, split_factor=3
        will make groupings of features of size 3 for the histograms.
        Note: this will likely make synthetic data worse.
        defaults to None
    :type split_factor: int, optional
    :param marginal_width: MWEM by default will create cuboids to measure
        marginals, and will use a heuristic to determine the maximum width
        of the marginals.  This parameter allows you to specify that MWEM
        should always measure marginals up to width marginal_width.
    :type marginal_width: int, optional
    :param add_ranges: In addition to measuring cuboids, MWEM can measure
        randomly generated range queries.  Range queries work well on columns
        that are binned from continuous values.
    :type add_range: bool, optional
    :param measure_only: MWEM operates by spending some privacy budget to select
        the best query.  This parameter allows you to specify that MWEM should
        uniformly measure all queries, and not spend any privacy budget on the
        query selection.  This is useful in limited cases.  Allowing MWEM to select
        will usually work best.
    :type measure_only: bool, optional
    :param max_retries_exp_mechanism: In each iteration, MWEM tries to select
        a poorly-peforming query that hasn't yet been measured.  If it fails,
        it will select one of the remaining queries uniformly at random.
    :type max_retries_exp_mechanism: int, optional
    :param mult_weights_iterations: Number of iterations of multiplicative weights,
        per iteration of MWEM, defaults to 20
    :type mult_weights_iterations: int, optional
    :param verbose: Set to True to print debug information.
    :type verbose: bool, optional
    """
    def __init__(
        self,
        epsilon=3.0,
        *ignore,
        q_count=None,
        iterations=None,
        splits=[],
        split_factor=None,
        marginal_width=None,
        add_ranges=False,
        measure_only=False,
        max_retries_exp_mechanism=10,
        mult_weights_iterations=20,
        verbose=False
    ):
        self.epsilon = epsilon
        self.q_count = q_count
        self.iterations = iterations
        self.mult_weights_iterations = mult_weights_iterations
        self.add_ranges = add_ranges
        self.measure_only = measure_only
        self.synthetic_data = None
        self.data_bins = None
        self.real_data = None
        self.splits = splits
        self.split_factor = split_factor
        self.mins_maxes = {}
        self.scale = {}
        self.marginal_width = marginal_width
        self.debug = verbose

        # Pandas check
        self.pandas = False
        self.pd_cols = None
        self.pd_index = None

        # Query trackers
        self.q_values = None
        self.max_retries_exp_mechanism = max_retries_exp_mechanism
        self.accountant = []

    @property
    def spent(self):
        return sum([a for a in self.accountant])

    def fit(
            self,
            data, *ignore,
            transformer=None,
            categorical_columns=None,
            ordinal_columns=None,
            continuous_columns=None,
            preprocessor_eps=0.0,
            nullable=False):
        """
        Follows sdgym schema to be compatible with their benchmark system.

        :param data: Dataset to use as basis for synthetic data
        :type data: np.ndarray
        :return: synthetic data, real data histograms
        :rtype: np.ndarray
        """

        train_data = self._get_train_data(
            data,
            style='cube',
            transformer=transformer,
            categorical_columns=categorical_columns,
            ordinal_columns=ordinal_columns,
            continuous_columns=continuous_columns,
            nullable=nullable,
            preprocessor_eps=preprocessor_eps
        )

        data = train_data

        if isinstance(data, np.ndarray):
            self.data = data.copy()
        elif isinstance(data, pd.DataFrame):
            self.pandas = True
            for col in data.columns:
                data[col] = pd.to_numeric(data[col], errors="ignore")
            self.data = data.to_numpy().copy()
            self.pd_cols = data.columns
            self.pd_index = data.index
        elif isinstance(data, list):
            self.data = np.array(data)
        else:
            raise ValueError("Data must be a list of tuples, a numpy array, or a pandas dataframe.")
        if self.split_factor is not None and self.splits == []:
            self.splits = self._generate_splits(self.data.T.shape[0], self.split_factor)
        elif self.split_factor is None and self.splits == []:
            # Set split factor to default to shape[1]
            self.split_factor = self.data.shape[1]
            self.splits = self._generate_splits(self.data.T.shape[0], self.split_factor)

        self.splits = np.array(self.splits)
        if self.splits.size == 0:
            self.histograms = self._histogram_from_data_attributes(
                self.data, [np.arange(self.data.shape[1])]
            )
        else:
            self.histograms = self._histogram_from_data_attributes(self.data, self.splits)
        if self.debug:
            print(f"Processing {len(self.histograms)} histograms")
            print()

        # load queries into each split
        n_histograms = len(self.histograms)
        for idx, h in enumerate(self.histograms):
            marginal_width = self.marginal_width
            scale = np.sum(h.data)
            iterations = self.iterations
            if iterations is None:
                iterations = int(np.ceil(np.log10(scale))) * 10

            if marginal_width is None:
                if scale > 10_000:
                    marginal_width = 3
                else:
                    marginal_width = 2

            h.add_cuboid_queries(marginal_width)

            q_count = self.q_count if self.q_count is not None else 2 * iterations

            while h.n_queries < q_count:
                if self.add_ranges:
                    h.add_arbitrary_queries(q_count - h.n_queries)
                else:
                    h.add_cuboid_queries(marginal_width)
            if h.n_queries > q_count:
                h.queries = np.random.choice(h.queries, q_count, replace=False)

            meas_eps = self.epsilon / 2.0 / n_histograms / iterations
            meas_bounds = -np.log(2*(0.05)) * (1 / meas_eps)  # 95% confidence
            h.iterations = iterations
            h.meas_eps = meas_eps
            h.meas_bound = meas_bounds

            if self.debug:
                print(f"Histogram #{idx} split: {h.split}")
                print(f"Columns: {h.n_cols}")
                print(f"Dimensionality: {h.dimensionality:,}")
                print(f"Cuboids possible: {h.n_cuboids}")
                print(f"1-2-way cuboids possible: {math.comb(h.n_cols, 2) + h.n_cols}")
                print(f"Fitting for {h.iterations} iterations")
                print(f"Number of queries: {h.n_queries}")
                print(f"Number of slices in queries: {h.n_slices}")
                print(f"Per-Measure Epsilon: {h.meas_eps:.3f}")
                print(f"Measurement Error: {h.meas_bound:.2f}")
                print()

        self.max_iterations = max([h.iterations for h in self.histograms])

        # Run the algorithm
        self.synthetic_histograms = self.mwem()

    def sample(self, samples):
        """
        Creates samples from the histogram data.
        Follows sdgym schema to be compatible with their benchmark system.
        NOTE: We are sampling from each split dimensional
        group as though they are *independent* from one another.
        We have essentially created len(splits) DP histograms as
        if they are separate databases, and combine the results into
        a single sample.

        :param samples: Number of samples to generate
        :type samples: int
        :return: N samples
        :rtype: list(np.ndarray)
        """
        synthesized_columns = ()
        first = True
        for fake, _, split in self.synthetic_histograms:
            s = []
            fake_indices = np.arange(len(np.ravel(fake)))
            fake_distribution = np.ravel(fake)
            norm = np.sum(fake)
            for _ in range(samples):
                s.append(np.random.choice(fake_indices, p=(fake_distribution / norm)))
            s_unraveled = []
            for ind in s:
                s_unraveled.append(np.unravel_index(ind, fake.shape))
            # Here we make scale adjustments to match the original
            # data
            np_unraveled = np.array(s_unraveled)
            for i in range(np_unraveled.shape[-1]):
                min_c, max_c = self.mins_maxes[str(split[i])]
                # TODO: Deal with the 0 edge case when scaling
                # i.e. scale factor * 0th bin is 0,
                # but should still scale appropriately
                np_unraveled[:, i] = np_unraveled[:, i] * self.scale[str(split[i])]
                np_unraveled[:, i] = np_unraveled[:, i] + min_c
            if first:
                synthesized_columns = np_unraveled
                first = False
            else:
                synthesized_columns = np.hstack((synthesized_columns, np_unraveled))
        # Recombine the independent distributions into a single dataset
        combined = synthesized_columns
        # Reorder the columns to mirror their original order
        r = self._reorder(self.splits)
        return self._transformer.inverse_transform(combined[:, r])

    def mwem(self):
        """
        Runner for the mwem algorithm.
        Initializes the synthetic histogram, and updates it
        for up to self.max_iterations using the exponential mechanism and
        multiplicative weights. Draws from the initialized query store
        for measurements.

        :return: synth_hist, self.histogram - synth_hist is the
            synthetic data histogram, self.histogram is original histo
        :rtype: np.ndarray, np.ndarray
        """
        a_values = []
        for idx, h in enumerate(self.histograms):
            hist = h.data
            dimensions = h.dimensions
            split = h.split
            queries = h.queries
            synth_hist = self._initialize_a(hist, dimensions)
            measurements = {}
            # NOTE: Here we perform a privacy check,
            # because if the histogram dimensions are
            # greater than the iterations, this can be
            # a big privacy risk (the sample queries will
            # otherwise be able to match the actual
            # distribution)
            # This usually occurs with a split factor of 1,
            # so that each attribute is independent of the other
            flat_dim = h.dimensionality
            iterations = h.iterations
            if 2 * flat_dim <= iterations:
                warnings.warn(
                    "Flattened dimensionality of synthetic histogram is less than"
                    + " the number of iterations. This is a privacy risk."
                    + " Consider increasing your split_factor (especially if it is 1), "
                    + "or decreasing the number of iterations. "
                    + "Dim: " + str(flat_dim) + " Split: " + str(split),
                    Warning,
                )

            eps = h.meas_eps if not self.measure_only else 2 * h.meas_eps
            for i in range(iterations):
                qi = self._exponential_mechanism(
                    hist, synth_hist, queries, eps, measurements, i
                )

                for query in queries[qi].queries:
                    assert (isinstance(query, Query))
                    actual = query.evaluate(hist)
                    lap = self._laplace(1.0/eps)
                    if qi in measurements:
                        measurements[qi].append(actual + lap)
                    else:
                        measurements[qi] = [actual + lap]
                self.accountant.append(eps)
                synth_hist = self._multiplicative_weights(
                    synth_hist, queries, measurements, hist, self.mult_weights_iterations
                )
            a_values.append((synth_hist, hist, split))
        return a_values

    def _initialize_a(self, histogram, dimensions):
        """
        Initializes a uniform distribution histogram from
        the given histogram with dimensions

        :param histogram: Reference histogram
        :type histogram: np.ndarray
        :param dimensions: Reference dimensions
        :type dimensions: np.ndarray
        :return: New histogram, uniformly distributed according to
            reference histogram
        :rtype: np.ndarray
        """
        n = np.sum(histogram)
        value = n / np.prod(dimensions)
        synth_hist = np.zeros_like(histogram)
        synth_hist += value
        return synth_hist

    def _histogram_from_data_attributes(self, data, splits=[]):
        """
        Create a histogram from given data

        :param data: Reference histogram
        :type data: np.ndarray
        :return: Histogram over given data, dimensions,
            bins created (output of np.histogramdd)
        :rtype: np.ndarray, np.shape, np.ndarray
        """
        histograms = []
        for split in splits:
            split_data = data[:, split]
            mins_data = []
            maxs_data = []
            dims_sizes = []
            # Transpose for column wise iteration
            for i, column in enumerate(split_data.T):
                min_c = min(column)
                max_c = max(column)
                # TODO: just use integer bins and assume 0 base
                mins_data.append(min_c)
                maxs_data.append(max_c)
                # Dimension size (number of bins)
                bin_count = int(max_c - min_c + 1)
                # Here we track the min and max for the column,
                # for sampling
                self.mins_maxes[str(split[i])] = (min_c, max_c)
                self.scale[str(split[i])] = 1
                dims_sizes.append(bin_count)
            # Produce an N,D dimensional histogram, where
            # we pre-specify the bin sizes to correspond with
            # our ranges above
            if any([a > 0 for a in mins_data]):
                warnings.warn("Data should be preprocessed to have 0 based indices.")
            dimensionality = np.product(dims_sizes)
            if dimensionality > 1e8:
                warnings.warn(f"Dimensionality of histogram is {dimensionality:,}, consider using splits.")
            histogram, bins = np.histogramdd(split_data, bins=dims_sizes)
            # Return histogram, dimensions
            h = Histogram(histogram, dims_sizes, bins, split)
            histograms.append(h)
        return histograms

    def _exponential_mechanism(self, hist, synth_hist, queries, eps, measurements, iteration):
        """
        Refer to paper for in depth description of
        Exponential Mechanism.
        Parametrized with epsilon value epsilon/(2 * iterations)

        :param hist: Basis histogram
        :type hist: np.ndarray
        :param synth_hist: Synthetic histogram
        :type synth_hist: np.ndarray
        :param queries: Queries to draw from
        :type queries: list
        :param eps: Budget
        :type eps: float
        :return: # of errors
        :rtype: int
        """
        errors = [queries[i].error(hist, synth_hist) * (eps / 2.0)
                  for i in range(len(queries))
                  ]
        maxi = max(errors)
        mean_err = np.mean(errors)
        exp_errors = [math.exp(errors[i] - maxi) for i in range(len(errors))]

        count_retries = 0
        qi = None
        while qi is None or qi in measurements:
            if self.measure_only or count_retries > self.max_retries_exp_mechanism:
                # grab one uniformly random, wastes epsilon, but is safe
                options = [i for i in range(len(queries))]
                options = [i for i in options if i not in measurements]
                qi = np.random.choice(options)
            else:
                r = random.random()
                e_s = sum(exp_errors)
                c = 0
                qi = len(exp_errors) - 1
                for i in range(len(exp_errors)):
                    c += exp_errors[i]
                    if c > r * e_s:
                        qi = i
                        break
                count_retries += 1

        if self.debug:
            log = int(np.floor(np.log10(self.max_iterations)))
            skip = 1 if log < 2 else 10 ** (log - 1)
            if iteration % skip == 0:
                print(f"[{iteration}] - Average error: {mean_err:.3f}. Selected {len(queries[qi].queries)} slices")
        if not self.measure_only:
            self.accountant.append(eps)
        return qi

    def _multiplicative_weights(self, synth_hist, queries, m, hist, iterate):
        """
        Multiplicative weights update algorithm,
        used to boost the synthetic data accuracy given measurements m.
        Run for iterate times

        :param synth_hist: Synthetic histogram
        :type synth_hist: np.ndarray
        :param queries: Queries to draw from
        :type queries: list
        :param m: Measurements taken from real data for each qi query
        :type m: dict
        :param hist: Basis histogram
        :type hist: np.ndarray
        :param iterate: Number of iterations to run mult weights
        :type iterate: iterate
        :return: synth_hist
        :rtype: np.ndarray
        """
        sum_a = np.sum(synth_hist)
        for _ in range(iterate):
            for qi in m:
                measurements = m[qi]
                queries_list = queries[qi].queries
                assert (len(measurements) == len(queries_list))

                for measurement, query in zip(measurements, queries_list):
                    error = measurement - query.evaluate(synth_hist)
                    query_update = query.mask(synth_hist)

                    # Apply the update
                    a_multiplier = np.exp(query_update * error / (2.0 * sum_a))
                    a_multiplier[a_multiplier == 0.0] = 1.0
                    synth_hist = synth_hist * a_multiplier
                    # Normalize again
                    count_a = np.sum(synth_hist)
                    synth_hist = synth_hist * (sum_a / count_a)
        return synth_hist

    def _reorder(self, splits):
        """
        Given an array of dimensionality splits (column indices)
        returns the corresponding reorder array (indices to return
        columns to original order)
        Example:
        original = [[1, 2, 3, 4, 5, 6],
        [ 6,  7,  8,  9, 10, 11]]

        splits = [[1,3,4],[0,2,5]]

        mod_data = [[2 4 5 1 3 6]
                [ 7  9 10  6  8 11]]

        reorder = [3 0 4 1 2 5]

        :param splits: 2d list with splits (column indices)
        :type splits: array of arrays
        :return: 2d list with splits (column indices)
        :rtype: array of arrays
        """
        flat = np.concatenate(np.asarray(splits)).ravel()
        reordered = np.zeros(len(flat))
        for i, ind in enumerate(flat):
            reordered[ind] = i
        return reordered.astype(int)

    def _generate_splits(self, n_dim, factor):
        """
        If user specifies, do the work and figure out how to divide the dimensions
        into even splits to speed up MWEM
        Last split will contain leftovers <= sizeof(factor)

        :param n_dim: Total # of dimensions
        :type n_dim: int
        :param factor: Desired size of the splits
        :type factor: int
        :return: Splits
        :rtype: np.array(np.array(),...)
        """
        # Columns indices
        indices = np.arange(n_dim)

        # Split intelligently
        fits = int((np.floor(len(indices) / factor)) * factor)
        even_inds = indices[:fits].copy().reshape((int(len(indices) / factor), factor))
        s1 = even_inds.tolist()
        if indices[fits:].size != 0:
            s1.append(indices[fits:])
        s2 = [np.array(l_val) for l_val in s1]
        return np.array(s2)

    def _laplace(self, sigma):
        """
        Laplace mechanism

        :param sigma: Laplace scale param sigma
        :type sigma: float
        :return: Random value from laplace distribution [-1,1]
        :rtype: float
        """
        return sigma * np.log(random.random()) * np.random.choice([-1, 1])



import pandas as pd
import numpy as np
import warnings

from snsql.sql.odometer import OdometerHeterogeneous
from snsql.sql.privacy import Privacy

from .anonymization import AnonymizationTransformer
from .drop import DropTransformer
from snsynth.transform.type_map import TypeMap

class TableTransformer:
    """Transforms a table of data.

    :param transformers: a list of transformers, one per column
    :param odometer: an optional odometer to use to track privacy spent when fitting the data
    """
    def __init__(self, transformers=[], *ignore, odometer=None):
        # one transformer per input column
        self.transformers = transformers
        if self.fit_complete:
            self.output_width = sum([t.output_width for t in self.transformers])
        else:
            self.output_width = 0
        if odometer is None:
            self.odometer = OdometerHeterogeneous(privacy=Privacy())
        else:
            self.odometer = odometer

        self._columns = None # will be automatically set if pandas
        self._dtype = None # will be automatically set if numpy
        self._dropped_column_indices = set()  # used if a pd.DataFrame and at least one DropTransformer are provided

    @property
    def fit_complete(self):
        """Returns True if the transformer has been fit."""
        return all([t.fit_complete for t in self.transformers])
    @property
    def needs_epsilon(self):
        """Returns True if the transformer needs to spend privacy budget when fitting."""
        return any([t.needs_epsilon for t in self.transformers])
    @property
    def cardinality(self):
        """Returns the cardinality of each output column.  Returns None for continuous columns."""
        cards = []
        for t in self.transformers:
            if isinstance(t, DropTransformer) or isinstance(t, AnonymizationTransformer):
                continue # don't include cardinality for the unbounded transformers
            for c in t.cardinality:
                cards.append(c)
        return cards
    def allocate_privacy_budget(self, epsilon, odometer):
        n_with_epsilon = sum([1 for t in self.transformers if t.needs_epsilon])
        if n_with_epsilon == 0:
            return
        else:
            for transformer in self.transformers:
                if transformer.needs_epsilon:
                    transformer.allocate_privacy_budget(epsilon / n_with_epsilon, odometer)
    def fit(self, data, *ignore, epsilon=None):
        """Fits the transformer to the data.

        :param data: a table represented as a list of tuples, a numpy.ndarray, or a pandas DataFrame
        :param epsilon: the privacy budget to spend fitting the data
        """
        if self.transformers == []:
            self._fit_finish()
        if epsilon is not None and epsilon > 0.0:
            self.allocate_privacy_budget(epsilon, self.odometer)
        if isinstance(data, pd.DataFrame):
            self._columns = list(data.columns)
            data = [tuple([c for c in t[1:]]) for t in data.itertuples()]
        self._dropped_column_indices = set()
        for t in self.transformers:
            t._clear_fit()
        for row in data:
            self._fit(row)
        for t in self.transformers:
            t._fit_finish()
        self._fit_finish()
        if self.output_width == 0:
            warnings.warn("No columns were selected for output.  This may be because all columns were anonymized.")
    def _fit(self, row):
        for v, t in zip(row, self.transformers):
            t._fit(v)
    def _fit_finish(self):
        self.output_width = sum([t.output_width for t in self.transformers])
    def transform(self, data):
        """Transforms the data.

        :param data: tabular data to transform
        :type data: a list of tuples, a numpy.ndarray, or a pandas DataFrame
        :returns: the transformed data
        :rtype: a list of tuples
        """
        # always returns a list of tuples, except in null case
        if self.transformers == []:
            return data
        if isinstance(data, pd.DataFrame):
            if self._columns is not None:
                # check that columns match
                columns = list(data.columns)
                if len(columns) != len(self._columns):
                    raise ValueError(f"Wrong number of columns: got {len(columns)}, expected {len(self._columns)}")
                if not(all([a == b for a, b in zip(columns, self._columns)])):
                    warnings.warn(f"Columns of data do not match columns of transformer: {columns} vs {self._columns}")
            self._columns = data.columns
            data = [tuple([c for c in t[1:]]) for t in data.itertuples()]
        elif isinstance(data, np.ndarray):
            self._dtype = data.dtype
            if len(data.shape) != 2:
                raise ValueError(f"Data must be a 2D array, got shape {data.shape}")
            if data.shape[1] != len(self.transformers):
                raise ValueError(f"Data must have {len(self.transformers)} columns, got {data.shape[1]}")
            data = [tuple([c for c in t]) for t in data]
        return [self._transform(row) for row in data]
    def _transform(self, row):
        out_row = []
        for v, t in zip(row, self.transformers):
            if isinstance(t, AnonymizationTransformer) and not t.fake_inbound:
                pass  # don't include any values if we wish to anonymize with inverse transformation
            elif isinstance(t, DropTransformer):
                pass  # don't include any values from DropTransformer
            elif t.output_width == 1:
                out_row.append(t._transform(v))
            else:
                for out_v in t._transform(v):
                    out_row.append(out_v)
        return tuple(out_row)
    def fit_transform(self, data, *ignore, epsilon=None):
        """Fits the transformer to the data, then transforms.

        :param data: tabular data to transform
        :type data: a list of tuples, a numpy.ndarray, or a pandas DataFrame
        :param epsilon: the privacy budget to spend fitting the data
        :type epsilon: float, optional
        :returns: the transformed data
        :rtype: a list of tuples
        """
        self.fit(data, epsilon=epsilon)
        return self.transform(data)
    def inverse_transform(self, data):
        if self.transformers == []:
            return data
        transformed = [self._inverse_transform(row) for row in data]
        if self._columns is not None:
            columns = [col for i, col in enumerate(self._columns) if i not in self._dropped_column_indices]
            return pd.DataFrame(transformed, columns=columns)
        elif self._dtype is not None:
            return np.array(transformed, dtype=self._dtype)
        else:
            return transformed
    def _inverse_transform(self, row):
        if len(row) != self.output_width:
            raise ValueError(f"Row has wrong length: got {len(row)}, expected {self.output_width}")
        out_row = []
        row = list(row)
        for i, t in enumerate(self.transformers):
            if isinstance(t, DropTransformer): # don't include None values from DropTransformer
                self._dropped_column_indices.add(i) # and mark column index as dropped
                continue

            if t.output_width == 1:
                v = row.pop(0)
            else:
                v = tuple([row.pop(0) for _ in range(t.output_width)])
            out_row.append(t._inverse_transform(v))
        return tuple(out_row)

    # factory methods
    @classmethod
    def from_column_names(cls, column_names, style='gan', *ignore, nullable=False, categorical_columns=[], ordinal_columns=[], continuous_columns=[], special_types={}, constraints=None):
        transformers = TypeMap.get_transformers(column_names, style=style, nullable=nullable, categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, special_types=special_types, constraints=constraints)
        return cls(transformers)
    @classmethod
    def from_pandas(cls, df, style='gan', *ignore, nullable=False, categorical_columns=[], ordinal_columns=[], continuous_columns=[], special_types={}, constraints=None):
        return cls.from_column_names(df.columns, style=style, nullable=nullable, categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, special_types=special_types, constraints=constraints)
    @classmethod
    def from_list(cls, data, style='gan', *ignore, nullable=False, header=False, categorical_columns=[], ordinal_columns=[], continuous_columns=[], special_types={}, constraints=None):
        if not header:
            column_names = list(range(len(data[0])))
        else:
            column_names = data[0]
        return cls.from_column_names(column_names, style=style, nullable=nullable, categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, special_types=special_types, constraints=constraints)
    @classmethod
    def from_numpy(cls, data, style='gan', *ignore, nullable=False, categorical_columns=[], ordinal_columns=[], continuous_columns=[], special_types={}, constraints=None):
        column_names = list(range(len(data[0])))
        return cls.from_column_names(column_names, style=style, nullable=nullable, categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, special_types=special_types, constraints=constraints)
    @classmethod
    def create(cls, data, style='gan', *ignore, nullable=False, categorical_columns=[], ordinal_columns=[], continuous_columns=[], special_types={}, constraints=None):
        """
        Creates a transformer for the given data. Infers all columns if the provided lists are empty.
        Columns that are referenced in a constraint will be excluded from the inference.

        :param data: The private data to construct a transformer for.
        :type data: pd.DataFrame, np.ndarray, or list of tuples
        :param style: The style influences the choice of ColumnTransformers. Can either be 'gan' or 'cube'.
            Defaults to 'gan' which results in one-hot style.
        :type style: string, optional
        :param nullable: Whether to allow null values in the data. This is used as a hint when inferring transformers.
            Defaults to False.
        :type nullable: bool, optional
        :param categorical_columns: List of column names or indixes to be treated as categorical columns, used as hint.
        :type categorical_columns: list[], optional
        :param ordinal_columns: List of column names or indices to be treated as ordinal columns, used as hint.
        :type ordinal_columns: list[], optional
        :param continuous_columns: List of column names or indices to be treated as continuous columns, used as hint.
        :type continuous_columns: list[], optional
        :param constraints: Dictionary that maps from column names or indixes to constraints.
            There are multiple ways to specify a constraint.
            It can be a ``ColumnTransformer`` object, type or class name.
            Another possiblity is the string keyword 'drop' which enforces a ``DropTransformer``.
            Also, a string alias for any of the lists like 'categorical' can be provided.
            All other values e.g. a callable or Faker method will be passed into an ``AnonymizationTransformer``.
        :type constraints: dict, optional
        :returns: The transformer object
        :rtype: TableTransformer
        """
        if categorical_columns is None:
            categorical_columns = []
        if ordinal_columns is None:
            ordinal_columns = []
        if continuous_columns is None:
            continuous_columns = []
        if special_types is None:
            special_types = {}

        excluded_columns = None
        if constraints is not None: # exclude columns with constraint from inference
            if not isinstance(constraints, dict):
                raise ValueError("Provided `constraints` is invalid. Must be a dictionary or None.")
            excluded_columns = set(constraints.keys())

        if len(continuous_columns) + len(ordinal_columns) + len(categorical_columns) == 0:
            inferred = TypeMap.infer_column_types(data, excluded_columns=excluded_columns)
            categorical_columns = inferred['categorical_columns']
            ordinal_columns = inferred['ordinal_columns']
            continuous_columns = inferred['continuous_columns']
            special_types = dict(zip(inferred['columns'], inferred['pii']))
            if not nullable:
                nullable = len(inferred['nullable_columns']) > 0
        all_specified = list(categorical_columns) + list(ordinal_columns) + list(continuous_columns)
        all_numeric = all([isinstance(c, int) for c in all_specified])
        if isinstance(data, pd.DataFrame):
            return cls.from_pandas(data, style=style, nullable=nullable, categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, special_types=special_types, constraints=constraints)
        elif isinstance(data, np.ndarray):
            return cls.from_numpy(data, style=style, nullable=nullable, categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, special_types=special_types, constraints=constraints)
        elif isinstance(data, list):
            return cls.from_list(data, style=style, nullable=nullable, header=(not all_numeric), categorical_columns=categorical_columns, ordinal_columns=ordinal_columns, continuous_columns=continuous_columns, special_types=special_types, constraints=constraints)
        else:
            raise ValueError(f"Unknown data type: {type(data)}")

class NoTransformer(TableTransformer):
    """A pass-through table transformer that does nothing.  Note that the ``transform`` and ``inverse_transform`` methods
    will simply return the data that is passed in, rather than transforming to and from a list of tuples.  This transformer
    is suitable when you know that your input data is exactly what is needed for a specific synthesizer, and
    you want to skip all pre-processing steps.  If you want a passthrough transformer that is slightly more
    adaptable to multiple synthesizers, you can make a new ``TableTransformer`` with a list of ``IdentityTransformer``
    column transformers.
    """
    def __init__(self, *ignore):
        super().__init__([])
